//
// Functions related with calculating criticality values
// 
// O.S. 03/04/2002
//

#include <stdlib.h>
#include <cmath>
#include <deque>
#include <functional>
#include <algorithm>
#include <string>

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include "smodels.h"
#include "print.h"
#include "atomrule.h"

using namespace std;

#define CRIT(x) ((x)->criticality)
#define NEGCRIT(x) ((x)->negcriticality)

bool sort_op(Atom *a1, Atom *a2);
bool sort_op_hard(Atom *a1, Atom *a2);
void print_all_scores(Program *);

/*
int user_command(Smodels *s)
{
    char c[80];
    ifstream *fb = new ifstream("/dev/pts/0");

    cin = *fb;

    while (1)
    {	
	cout << endl;
	cout << "(n):Next Step (c):Choose (x):Choose Now (q):Quit (m>): Print model" << endl;
	cout << "(e):Expand (d>):Print dcl (s>):Print stack (h>):Print heuristic scores";
	cout << endl;
	cout << "(l):Toggle lookahead flag (u>):Print uncovered atoms (i):Print heuristic hi" << endl;
	cout << "(2):Find conflicting atom (r>):Print criticalities (o>):Print operator list" << endl;
	cout << "? ";
    
	for ( ;; )
	{
	    cin >> c;
	    if (cin)
	    {
		cin.ignore(INT_MAX, '\n');
		break;
	    }
	    cin.clear();
	    cin.ignore(INT_MAX, '\n');
	    cout << "Error in input." << endl;
	}

	// Choose a literal interactively
    	if (c[0] == 'c')
    	{
	    // get the atom's index in atom array.
	    char pos_neg = 0;
	    long num = -1; 

	    for  (;;)
	    {
	    	cout << "Index of atom to add: " << flush;
	    	cin >> num;
		if (cin)
		{ 
	    	    cin.ignore(INT_MAX, '\n');
		    if (num >= 0 && num < s->atomtop)
		    	break;
		    continue;
		}
		cin.clear();
	        cin.ignore(INT_MAX, '\n');
	    }
				    
	    for  (;;)
	    {
	    	cout << "Negative or positive [n/p]: " << flush;
	    	cin >> pos_neg;
		if (cin)
		{
	    	    cin.ignore(INT_MAX, '\n');
		    if (pos_neg == 'p' || pos_neg == 'n')
		    	break;
		    continue;
		}
		cin.clear();
	    	cin.ignore(INT_MAX, '\n');
	    }

	    s->hi_index = num;
	    if (pos_neg == 'p')
	    {
	    	cout << "Will add " << s->atom[num]->atom_name() << " to B." << endl;
		s->hi_is_positive = true;
	    }
	    else
	    {
	    	cout << "Will add not " << s->atom[num]->atom_name() << " to B." << endl;
		s->hi_is_positive = false;
	    }
		
	}
	// Choose NOW!
	else if (c[0] == 'x')
	{
	    char name[50];
	    char pos_neg;

	    for  (;;)
	    {
	    	cout << "Name of atom to add: " << flush;
	    	cin >> name;
		if (cin)
		{ 
	    	    cin.ignore(INT_MAX, '\n');
		    //if (num >= 0 && num < s->atomtop)
		    	break;
		    //continue;
		}
		cin.clear();
	        cin.ignore(INT_MAX, '\n');
	    }

	    for  (;;)
	    {
	    	cout << "Negative or positive [n/p]: " << flush;
	    	cin >> pos_neg;
		if (cin)
		{
	    	    cin.ignore(INT_MAX, '\n');
		    if (pos_neg == 'p' || pos_neg == 'n')
		    	break;
		    continue;
		}
		cin.clear();
	    	cin.ignore(INT_MAX, '\n');
	    }

	    // Find the atom in atom array
	    long idx;
	    bool found = false;
	    for (idx = 0; idx < s->atomtop; idx++)
		if (strcmp(name, s->atom[idx]->atom_name()) == 0)
		{
		    found = true;
		    break;
		}
	    
 	    if (!found)
	    {
		cout << "Atom could not found." << endl;
	    }
	    else
	    { 
	    	s->hi_index = idx;
	    	if (pos_neg == 'p')
		    s->hi_is_positive = true;
	    	else
		    s->hi_is_positive = false;

	    	s->choose();
	    }
	}
    	// Go on execution with the next step
    	else if (c[0] == 'n')
    	{
	    return 2;
    	}
 	// search for a conflicting atom.
	else if (c[0] == '2')
	{
	    long i;
	    long unw_index = s->stack.top;
	    for (i=0; i<s->program.number_of_atoms; i++)
	    {
		if (s->criticality_list[i]->Bneg || s->criticality_list[i]->Bpos)
		    continue;
		long idx;
		for (idx=0; idx < s->atomtop; idx++)
		    if (s->atom[idx] == s->criticality_list[i])
			break;
		s->hi_index = idx;
		s->hi_is_positive = true;
		s->choose();
		s->expand();
		if (s->conflict())
		{
		    //s->unwind_to(unw_index);
		    //s->hi_is_positive = false;
		    //s->choose();
		    s->backtrack(false);
		    s->expand();
		    if (s->conflict())
		    {
			// FOUND !
			cout << "Atom " << s->criticality_list[i]->atom_name() <<
				" will cause conflict" << endl;
		        s->unwind_to(unw_index);
		    }
		    else
			s->unwind_to(unw_index);
		}
		else
		{
		    // unwind to beginning and continue
		    s->unwind_to(unw_index);
		    continue;
		}
	    }
	}
    	// Exit program
    	else if (c[0] == 'q')
    	{
	    cout << "Exiting..." << endl;
	    exit(0);  
	}
	// toggle lookahead flag
	else if (c[0] == 'l')
    	{
	    char t_f;
	    for  (;;)
	    {
	    	cout << "Lookahead [t/f]: " << flush;
	    	cin >> t_f;
		if (cin)
		{
	    	    cin.ignore(INT_MAX, '\n');
		    if (t_f == 't' || t_f == 'f')
		    	break;
		    continue;
		}
		cin.clear();
	    	cin.ignore(INT_MAX, '\n');
	    }
	    if (t_f == 't')
		return 10;
	    else
		return 11;
	}
    	// Expand
    	else if (c[0] == 'e')
    	{
	    s->expand();
	    if (s->conflict())
	    {
		cout << "Conflict_2 occured*" << endl;
		s->backtrack(false);
	    }
    	}
    	// Print heuristic hi (i.e. what will the smodels heuristic choose)
    	else if (c[0] == 'i')
    	{
	    cout << s->hi_index << ' ' << s->hi_is_positive << ' ';
	    if (s->hi_is_positive)
		cout << s->atom[s->hi_index]->atom_name() << endl;
	    else
		cout << " not " << s->atom[s->hi_index]->atom_name() << endl;
    	}
    	// Print dcl information
    	else if (c[0] == 'd')
    	{
	    if (c[1] == '>')
	    {
		ofstream *of = new ofstream(c+2);
		cout = *of;
		s->dcl.print();
		of->flush();

		ofstream *ofs = new ofstream();
		ofs->attach(1);
		cout = *ofs;

		delete of;
		cout << c+2 << " is written." << endl;
	    }
	    else
	    	s->dcl.print();  
    	}
    	// Print stack 
    	else if (c[0] == 's')
    	{
	    if (c[1] == '>')
	    {
		ofstream *of = new ofstream(c+2);
		cout = *of;
	    	s->printStack();  
		of->flush();

		ofstream *ofs = new ofstream();
		ofs->attach(1);
		cout = *ofs;

		delete of;
		cout << c+2 << " is written." << endl;
	    }
	    else
	    	s->printStack();  
    	}
    	// Print candidate model
    	else if (c[0] == 'm')
    	{
	    if (c[1] == '>')
	    {
		ofstream *of = new ofstream(c+2);
		cout = *of;
	    	s->print();  
		of->flush();

		ofstream *ofs = new ofstream();
		ofs->attach(1);
		cout = *ofs;

		delete of;
		cout << c+2 << " is written." << endl;
	    }
	    else
	    	s->print();  
    	}
    	// Print sorted heuristic scores
    	else if (c[0] == 'h')
    	{
	    if (c[1] == '>')
	    {
		ofstream *of = new ofstream(c+2);
		cout = *of;
	    	print_all_scores(&(s->program));  
		of->flush();

		ofstream *ofs = new ofstream();
		ofs->attach(1);
		cout = *ofs;

		delete of;
		cout << c+2 << " is written." << endl;
	    }
	    else
	    	print_all_scores(&(s->program));  
    	}
    	// Print uncovered atoms list (atom array index and name)
    	else if (c[0] == 'u')
    	{
	    if (c[1] == '>')
	    {
		ofstream *of = new ofstream(c+2);
		cout = *of;
		for (int i=0; i<s->atomtop; i++)
		    cout << i << '\t' << s->atom[i]->atom_name() << endl;
		of->flush();

		ofstream *ofs = new ofstream();
		ofs->attach(1);
		cout = *ofs;

		delete of;
		cout << c+2 << " is written." << endl;
	    }
	    else
		for (int i=0; i<s->atomtop; i++)
		    cout << i << '\t' << s->atom[i]->atom_name() << endl;
    	}
    	// Print criticalities
    	else if (c[0] == 'r')
    	{
	    if (c[1] == '>')
	    {
		ofstream *of = new ofstream(c+2);
		cout = *of;
	    	s->sortCriticalities();  
		s->printCriticalities();  
		of->flush();

		ofstream *ofs = new ofstream();
		ofs->attach(1);
		cout = *ofs;

		delete of;
		cout << c+2 << " is written." << endl;
	    }
	    else
	    {
	    	s->sortCriticalities();  
	    	s->printCriticalities();  
	    }
    	}
    	// Print all oplist
    	else if (c[0] == 'o')
    	{
	    if (c[1] == '>')
	    {
		ofstream *of = new ofstream(c+2);
		cout = *of;
	    	s->printAllOplist();  
		of->flush();

		ofstream *ofs = new ofstream();
		ofs->attach(1);
		cout = *ofs;

		delete of;
		cout << c+2 << " is written." << endl;
	    }
	    else
	    	s->printAllOplist();  
    	}
    	else if (c[0] == 0)	// error in input
    	{
	    cout << "Error in command." << endl;
    	}
    } // end while
    return 0;
}
*/

/*
int
Smodels::interactive_smodels (bool look, bool jump)
{
  if (look)
  {
    setup_with_lookahead ();
    PRINT_TRACE (cout << "<< setup_with_looakahead" << endl);
  }
  else
  {
    setup ();
    PRINT_TRACE (cout << "<< setup" << endl);
  }
  if (conflict ())
  {
    PRINT_TRACE (cout << "<< conflict" << endl);
    return 0;
  }
    
  PRINT_TRACE (cout << "<< conflict_1" << endl);
  //PRINT_TRACE (print_all_scores(&program));

  initCriticalities();
  calculateCritic(6);
  sortCriticalities();

  if (look && !covered ())
    {
      //PRINT_TRACE (print_all_scores(&program));
      heuristic ();
      PRINT_TRACE (cout << "<< heuristic" << endl);
      //PRINT_TRACE (print_all_scores(&program));
      int ret = user_command(this);
      if (ret == 2)
      {
	improve();	
        PRINT_TRACE (cout << "<< improve" << endl);
        choose ();
        PRINT_TRACE (cout << "<< choose" << endl);
      }
    }
  while (!fail)
    {
      PRINT_DCL (cout << "Smodels call" << endl;
		 dcl.print ());
      PRINT_BF (print ());
      PRINT_STACK (printStack ());
      PRINT_TRACE (cout << ">> expand_in_loop" << endl);
      expand ();
      PRINT_TRACE (cout << "<< expand" << endl);
      PRINT_BF (cout << "Expand" << endl);
      PRINT_DCL (dcl.print ());
      PRINT_BF (print ());
      PRINT_PROGRAM(printProgram ());
      PRINT_STACK (printStack ());
      if (conflict ())
      {
        PRINT_TRACE (cout << "<< conflict_2" << endl);
        user_command(this);
	backtrack (jump);
      }
      else if (covered ())
	{
	  answer_number++;
	  level = guesses;
	  cout << "Answer: " << answer_number << endl;
	  printAnswer ();
	  program.set_optimum ();
          user_command(this);
	  if (max_models && answer_number >= max_models)
	    return 1;
	  else
	    backtrack (jump);
	}
      else if (look)
      {
        PRINT_TRACE (cout << ">> lookahead_in_loop" << endl);
   	lookahead ();
        PRINT_TRACE (cout << "<< lookahead" << endl);
      }
      else
      {
        PRINT_TRACE (cout << ">> pick" << endl);
	pick ();
        PRINT_TRACE (cout << "<< pick" << endl);
      }
    }
  number_of_wrong_choices--;
  return 0;
}
*/

void
Smodels::generateOplist()
{
    for (Node *n = program.rules.head(); n; n=n->next)
    {
	// call generateOplist for all rules of the program.
        n->rule->generateOplist();
    }
}

void
Smodels::printAllOplist()
{
    for (Node *n = program.atoms.head(); n; n = n->next)
    {
	cout << n->atom->atom_name() << endl;
	n->atom->printOplist();
    }
}	

void
Smodels::calculateCritic(int iteration)
{
    //user_command(this);
    cout << "Calculating criticalities... " << flush;
    for (int i=1; i <= iteration; i++)
    {
	calculateNextCritic_2();
	cout << '[' << i << "] ";
	//user_command(this);
    }
    cout << endl;
    cout << "<< calculateCritic" << endl;
}

void
Smodels::calculateCritic_wAccuracy()
{
    //user_command(this);
    cout << "Calculating criticalities... " << flush;

    long i = 1;
    continue_iteration = true;
    //accuracy_digit = 2;
    while (continue_iteration)
    {
        continue_iteration = false;
	calculateNextCritic_2();
	cout << '[' << i++ << "] ";
    }

    cout << endl;
}

void
Smodels::printCriticalities()
{
    cout.setf(ios::showpoint);
    for (long i=0; i < program.number_of_atoms; i++)
    {
	// 28/05/2002 
	if (criticality_list[i] == 0)
	    break;
	cout << criticality_list[i]->atom_name() << "\t";
	cout << setprecision(6) << criticality_list[i]->criticality << ' ';
	cout << setprecision(6) << criticality_list[i]->negcriticality << '\t';
	cout << criticality_list[i]->posEffort << ' ';
	cout << criticality_list[i]->negEffort << '\t';
	for (long j=0; j<criticality_list[i]->number_of_choicerule; j++)
	{
	    cout << criticality_list[i]->catom_list[j].catom << ' ';
	    cout << criticality_list[i]->catom_list[j].ncatom << '\t';
	}
	cout << '\n';
    }
}

void
Smodels::initCriticalities()
{
    // first initialize the uncovered atoms
    for (int i=0; i<atomtop; i++)
    {
	assert((atom[i]->Bpos || atom[i]->Bneg) == false);

        atom[i]->init_criticality = 1;
        atom[i]->init_negcriticality = 1;

	atom[i]->criticality = atom[i]->init_criticality;
	atom[i]->next_criticality = atom[i]->criticality;
	atom[i]->negcriticality = atom[i]->init_negcriticality;
	atom[i]->next_negcriticality = atom[i]->negcriticality;

   	// generate catom_list;
	if (atom[i]->number_of_choicerule != 0)
	{
	    for (int j=0; j<atom[i]->number_of_choicerule; ++j)
	    {
		atom[i]->catom_list[j].catom = 1;
		atom[i]->catom_list[j].init_catom = 1;
		atom[i]->catom_list[j].next_catom = 1;
		atom[i]->catom_list[j].ncatom = 1;
		atom[i]->catom_list[j].init_ncatom = 1;
		atom[i]->catom_list[j].next_ncatom = 1;
	    }
	}
    }

    // now initialize the covered atoms.
    for (int i=atomtop; i<program.number_of_atoms; i++)
    {
	assert((atom[i]->Bpos || atom[i]->Bneg) == true);

        if (atom[i]->Bpos)
        {
	    atom[i]->init_criticality = 0;	// smaller than the min crit.
	    atom[i]->init_negcriticality = 1;
     	}
	else
	{
	    atom[i]->init_criticality = 1;
	    atom[i]->init_negcriticality = 0; // smaller than the min crit.
     	}

	atom[i]->criticality = atom[i]->init_criticality;
	atom[i]->next_criticality = atom[i]->criticality;
	atom[i]->negcriticality = atom[i]->init_negcriticality;
	atom[i]->next_negcriticality = atom[i]->negcriticality;

    	for (int j=0; j<atom[i]->number_of_choicerule; ++j)
	{
	    atom[i]->catom_list[j].catom = 1;
	    atom[i]->catom_list[j].init_catom = 1;
	    atom[i]->catom_list[j].next_catom = 1;
	    if (atom[i]->Bpos)
	    {
		atom[i]->catom_list[j].ncatom = 0;
	    	atom[i]->catom_list[j].init_ncatom = 0;
	    	atom[i]->catom_list[j].next_ncatom = 0;
	    }
	    else
	    {
	    	atom[i]->catom_list[j].ncatom = 1;
	    	atom[i]->catom_list[j].init_ncatom = 1;
	    	atom[i]->catom_list[j].next_ncatom = 1;
	    }
	}
    }
}

void
Smodels::calculateEfforts()
{
    for (long i=0; i < atomtop; i++)
    {
	Atom *a = atom[i];
	bool r;

	r = findEffortPos(a);
	assert(r == false);

	r = findEffortNeg(a);
	assert(r == false);
    }
}

void
Smodels::calculateNextCritic_2()
{

    for (Node *n = program.atoms.head(); n; n = n->next)
    {
	Atom *a = n->atom;

	if ((a->Bneg) || (a->Bpos))
	{
	    continue;	// if the atom or its negation is in the stable model
			// there is no need to compute the criticality
	}

	a->next_criticality = 1 / a->init_criticality;
	a->next_negcriticality = 0;

	for (OperatorNode *opn = a->oplist->head(); opn; opn = opn->next)
	{
	    double rule_criticality = 0;
	    double rule_crit_for_nota = 1;
	    bool rule_cant_fire = false;

	    for (OperatorAtom *opa = opn->op->head(); opa; opa = opa->next)
	    {
		if (opa->a == a)
		    continue;	// same atom in body
		if (opa->negated)
		{
		    if (opa->a->Bneg)
			continue;	// discard criticality of the atom
		    if (opa->a->Bpos)
		    {
			rule_cant_fire = true;
			break;		// discard all rule
		    }	
		    rule_criticality += opa->a->negcriticality;
		    rule_crit_for_nota *= opa->a->criticality;
		}
		else
		{
		    if (opa->a->Bpos)
			continue;	// discard criticality of the atom
		    if (opa->a->Bneg)
		    {
			rule_cant_fire = true;
			break;		// discard all rule
		    }	
		    rule_criticality += opa->a->criticality;
		    rule_crit_for_nota *= opa->a->negcriticality;
		}
	    }

	    if (rule_cant_fire)
	    {
		continue;
	    }
	
	    if (opn->op->choicerule)
	    {
		Choice_criticality *cc = &(a->catom_list[opn->op->catom_index]);

		// { not ca } :- a V ....
		cc->next_ncatom = 1 / cc->init_ncatom;
		cc->next_ncatom += 1 / (rule_crit_for_nota * a->criticality);
		cc->next_ncatom = 1 / cc->next_ncatom;

		// ca :- na (not a).
		cc->next_catom = 1 / cc->init_catom;
		cc->next_catom += 1 / (rule_criticality + a->negcriticality);
		cc->next_catom = 1 / cc->next_catom;

		// a :- nca, ...
		rule_criticality += cc->ncatom;
		rule_crit_for_nota *= cc->catom;

                cc->catom = cc->next_catom;
		cc->ncatom = cc->next_ncatom;
	    }

	    assert(rule_criticality != 0);
	    assert(opn->op->generatenot == false);

	    a->next_criticality += 1 / rule_criticality;
	    a->next_negcriticality += rule_crit_for_nota;
	}

        a->next_criticality = 1 / a->next_criticality;
	
	// calculate na (not a)
	a->next_negcriticality = (1/a->init_negcriticality)+(1/a->next_negcriticality);
	a->next_negcriticality = 1 / a->next_negcriticality; 

   	// 01/05/2002
 	// iteration with accuracy
	if (continue_iteration == false && accuracy_digit != -1)
	{
            float old_d = floor(a->criticality * pow(10.0,(double)accuracy_digit));
            float new_d = floor(a->next_criticality * pow(10.0,(double)accuracy_digit));

            if (old_d != new_d)
	    {
	        continue_iteration = true;
	    }

            old_d = floor(a->negcriticality * pow(10.0,(double)accuracy_digit));
            new_d = floor(a->next_negcriticality * pow(10.0,(double)accuracy_digit));

            if (old_d != new_d)
	    {
		continue_iteration = true;
	    }
/*
	    for (long x=1; x <= accuracy_digit; x++)
	    {
		int old_d = a->criticality * pow(10, x);
		int new_d = a->next_criticality * pow(10, x);
		if (x != 1)
		{
		    old_d -= pow(10,x-1);
		    new_d -= pow(10,x-1);
		}
		
		if (old_d != new_d)
		{
		    continue_iteration = true;
		    break;
		}
	
		old_d = a->negcriticality * pow(10,x);
		new_d = a->next_negcriticality * pow(10,x);
		if (x != 1)
		{
		    old_d -= pow(10,x-1);
		    new_d -= pow(10,x-1);
		}
		
		if (old_d != new_d) 
		{
		    continue_iteration = true;
		    break;
		}
	    }
*/
	}
    }

    // now assign the criticalities to newly calculated ones
    for (Node *n = program.atoms.head(); n; n=n->next)
    {
	n->atom->criticality = n->atom->next_criticality;
	n->atom->negcriticality = n->atom->next_negcriticality;
    }
}

void
Smodels::sortCriticalities_Hard()
{
    deque<Atom *> col;

    for (long i=0; i<program.number_of_atoms; ++i)
    {
	if (criticality_list[i] == 0) // NULL
	    break;
	if (criticality_list[i]->Bneg || criticality_list[i]->Bpos)
	    continue;
	else
	    col.insert(col.end(), criticality_list[i]);
    }

    stable_sort(col.begin(), col.end(), not2(ptr_fun(sort_op_hard)));

    //cout << "Number of uncovered atoms: " << col.size() << endl;
    long i;
    for (i=0; i<col.size(); ++i)
    {
	criticality_list[i] = col[i];
	criticality_list[i]->crit_index = i;
    }

    criticality_list[i] = 0;	// NULL
}

void
Smodels::sortCriticalities()
{
    deque<Atom *> col;

    for (long i=0; i<program.number_of_atoms; ++i)
    {
	if (criticality_list[i] == 0) // NULL
	    break;
	if (criticality_list[i]->Bneg || criticality_list[i]->Bpos)
	    continue;
	else
	    col.insert(col.end(), criticality_list[i]);
    }

    sort(col.begin(), col.end(), not2(ptr_fun(sort_op)));

    //cout << "Number of uncovered atoms: " << col.size() << endl;
    long i;
    for (i=0; i<col.size(); ++i)
    {
	criticality_list[i] = col[i];
	criticality_list[i]->crit_index = i;
    }

    criticality_list[i] = 0;	// NULL
}

void
Smodels::chooseCritical3()
{
    // clear conflict_2_has_occured
    conflict_2_has_occured = false;

    for (long i=next_to_choose; i<program.number_of_atoms; i++)
    {
	// 28/05/2002	criticality_list should not hold atoms in WFM
	if (criticality_list[i] == 0)
	    // finished list
	    break;

	// check if the literal is covered already
	if ((criticality_list[i]->Bneg) || (criticality_list[i]->Bpos))
	{
	    continue;
	}
	else
	{
	    long idx;

	    next_to_choose = i+1;

	    // first find the index of that atom in array atom
	    for (idx=0; idx<atomtop; ++idx)
		if (atom[idx] == criticality_list[i])
		    break;

	    hi_index = idx;
	    if (criticality_list[i]->posEffort >= 
		   criticality_list[i]->negEffort)
	    {
		hi_is_positive = true;
		criticality_list[i]->poscrit_chosen = true;
	    }
	    else
	    {
		hi_is_positive = false;
		criticality_list[i]->negcrit_chosen = true;
	    }

	    return;
	}
    }	// end of for

    next_to_choose = 0;
    cout << "End of criticality_list is reached." << endl;
    chooseCritical3();	// recursive call, we need for completeness of heu.
}
