//
// Abstraction hierarchy stuff
//
// O.S. 07/02/2002 	For now just an utility function to print the scores of
//			all atoms in descending order to stdout.
//

// 02/07/2002 Austin
#include <cstring>

#include <iostream.h>
#include <stdlib.h>
#include <limits.h>
#include "program.h"
#include "atomrule.h"

int init_sort_array(Program *p, Atom **atoms);

int compare_scores(const void *x, const void *y)
{
    Atom *a = (Atom *)x;
    Atom *b = (Atom *)y;
    unsigned long min_a, min_b, max_a, max_b;

    if (a->neg_tested && a->pos_tested)
    {
	if (a->negScore < a->posScore)
	{
	    min_a = a->negScore;
	    max_a = a->posScore;
	}
	else
	{
	    min_a = a->posScore;
	    max_a = a->negScore;
	}
    }
    else
    {
	if (b->neg_tested && b->pos_tested)
	    return 1; // b daha once gelmeli
	else
	    return 0; // esit
    }

    if (b->neg_tested && b->pos_tested)
    {
	if (b->negScore < b->posScore)
	{
	    min_b = b->negScore;
	    max_b = b->posScore;
	}
	else
	{
	    min_b = b->posScore;
	    max_b = b->negScore;
	}
	
	if (min_a > min_b || (min_a == min_b &&  max_a > max_b))
	    return -1; // a once gelmeli
	else
	    return 1; // b once gelmeli 
    }
    else
    {
	return -1; // a daha once gelmeli
    }
}

int compare_scores_v2(const void *x, const void *y)
{
    Atom *a = (Atom *)x;
    Atom *b = (Atom *)y;
    unsigned long min_a, min_b, max_a, max_b;

    if (a->negScore < a->posScore)
    {
	min_a = a->negScore;
	max_a = a->posScore;
    }
    else
    {
	min_a = a->posScore;
	max_a = a->negScore;
    }

    if (b->negScore < b->posScore)
    {
	min_b = b->negScore;
	max_b = b->posScore;
    }
    else
    {
	min_b = b->posScore;
	max_b = b->negScore;
    }

    if (min_a == ULONG_MAX || max_a == ULONG_MAX)
    {
	// b nin scorelari da ULONG_MAX ise 0 doner
    	if (min_b == ULONG_MAX || max_b == ULONG_MAX)
	    return 0;
	else
	    return -1; 
    }
    else
    {
	// b nin scorelari ULONG_MAX ise a daha once olmali
    	if (min_b == ULONG_MAX || max_b == ULONG_MAX)
	    return 1;
	else
	{
	    if (min_a > min_b || (min_a == min_b &&  max_a > max_b))
	    	return 1; // a once gelmeli
	    else
	    	return -1; // b once gelmeli 
	}
    }
}

void print_all_scores(Program *p)
{
    Atom *array[p->number_of_atoms];
    int result;

    if (init_sort_array(p, array))
	return;

    // sort the atoms array according to compare_scores function
    //qsort(array, p->number_of_atoms, sizeof(array[0]), &compare_scores_v2);
    for (long x=0; x<p->number_of_atoms; x++)
	for (long y=x+1; y<p->number_of_atoms; y++)
	{
	    result = compare_scores_v2(array[x], array[y]);
	    if (result == -1)
	    {
		Atom *tmp = array[x];
		array[x] = array[y];
		array[y] = tmp;
	    }
	}

    for(long i=0; i<p->number_of_atoms; i++)
    {
	cout << array[i]->atom_name();
	if (strlen(array[i]->atom_name()) >= 15)
	    cout << '\t';
	else
	    cout << "\t\t";
	cout << 'n' << (array[i]->neg_tested?'+':'-') << ' ';
	cout << 'p' << (array[i]->pos_tested?'+':'-') << ' ';
	if (array[i]->negScore < array[i]->posScore)
	{
	    cout << array[i]->negScore << ' ' << array[i]->posScore << " p" << endl;
	}
	else
	{
	    cout << array[i]->posScore << ' ' << array[i]->negScore << " n" << endl;
	}

	array[i] = 0;
    }
}

int init_sort_array(Program *p, Atom **atoms)
{
    long i=0;

    for (Node *n=p->atoms.head(); n; n=n->next)
    {
	atoms[i++] = n->atom;
    }

    if (i==p->number_of_atoms)
	return 0;
    else
	// something went wrong
	return 1;
}    
